/*
 * VirtuaColor Copyright (C) 2009 Mark Nudelman
 *
 * Use a separate desktop background color for each virtual desktop.
 */

#define VIRTUACOLOR_VERSION    "1.0"

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include "Messages.h" // VirtuaWin message definitions

typedef int DeskID;

/* Global variables */
static HWND   virtuaWin                   = 0;
static DeskID currDeskid                  = 0;
static char   configFilePrefix[MAX_PATH]  = "";

/* Lines in the config file. */
static const char versionLine[]  = "virtuacolor ";
static const char colorLine[]    = "color ";


static const char configMessage[] =
	"VirtuaColor maintains a separate background color for each desktop.\n\n"
	"Set the background color of each desktop as usual by\n"
	"right-clicking on the desktop and selecting Properties / Desktop.\n"
	"\n\n\n\n"
	"VirtuaColor Copyright (C) 2009 Mark Nudelman\n"
	"    Greenwood Software\n"
	"    www.greenwoodsoftware.com";

/* ----------------------------------------------------------------------
 * Get the name of the config file for a specific desktop.
 */
static const char* configFileName(DeskID deskid)
{
	static char configFileNameBuf[MAX_PATH];
    snprintf(configFileNameBuf, sizeof(configFileNameBuf),
			"%s.%d", configFilePrefix, (int) deskid);
	return configFileNameBuf;
}

/* ----------------------------------------------------------------------
 * Set a system color.
 */
static void SetSysColor(INT element, COLORREF color)
{
	INT elements[1] = { element };
	COLORREF colors[1] = { color };
	SetSysColors(1, elements, colors);
}

/* ----------------------------------------------------------------------
 * Save current desktop color to a config file.
 */
static void saveColor()
{
	DWORD color;
    FILE* configFile;
    
    if (currDeskid <= 0)
        return;
	color = GetSysColor(COLOR_BACKGROUND);
    if ((configFile = fopen(configFileName(currDeskid), "w")) == NULL)
		return;
	fprintf(configFile, "%s 1\n%s %lx\n",
			versionLine, colorLine, (unsigned long) color);
	fclose(configFile);
}

/* ----------------------------------------------------------------------
 * Load current desktop color from a config file.
 */
static void loadColor()
{
	DWORD color = ~(DWORD)0;
    FILE* configFile;
	long version = 0;
	char line[256];

    if (currDeskid <= 0)
        return;
    if ((configFile = fopen(configFileName(currDeskid), "r")) == NULL)
		return;
	/* Read each line from the config file. */
	while (fgets(line, sizeof(line), configFile) != NULL) {
		const char* word = line + strspn(line, " ");
		if (word[0] == '#')
			/* comment */ ;
		else if (strncmp(word, versionLine, strlen(versionLine)) == 0)
			version = strtol(word + strlen(versionLine), NULL, 0);
		else if (strncmp(word, colorLine, strlen(colorLine)) == 0)
			color = strtoul(word + strlen(colorLine), NULL, 16);
	}
	fclose(configFile);
	if (!version) {
		/* Config file was missing the version line. */
		return;
	}
	SetSysColor(COLOR_BACKGROUND, color);
}

/* ----------------------------------------------------------------------
 * Switch to a new desktop.
 */
static void switchDesk(DeskID deskid)
{
    if (deskid <= 0 || deskid == currDeskid)
		return;
	saveColor();
	currDeskid = deskid;
	loadColor();
}

/* ----------------------------------------------------------------------
 */
static void initialize()
{
	if (configFilePrefix[0])
		return; /* already initialized */
	SendMessage(virtuaWin, VW_USERAPPPATH, 0, 0);
	if (configFilePrefix[0])
		return;
	MessageBox(virtuaWin, "cannot find UserApp path",
			"VirtuaColor Error", MB_ICONSTOP);
	exit(1);
}

/* ----------------------------------------------------------------------
 * Handle messages, mostly from VirtuaWin.
 */
static LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) 
{
    switch (msg)
    {
    case MOD_INIT: {
		DeskID deskid;
        virtuaWin = (HWND) wParam;
		initialize();
		deskid = SendMessage(virtuaWin, VW_CURDESK, 0, 0);
        switchDesk(deskid);
        break;
	}
    case WM_COPYDATA: {
		COPYDATASTRUCT* cds = (COPYDATASTRUCT*) lParam;
        if (cds->dwData == 0 - VW_USERAPPPATH) {
			snprintf(configFilePrefix, sizeof(configFilePrefix),
					"%s/VirtuaColor", (LPSTR)(cds->lpData));
            return TRUE;
        }
        break;
	}
    case MOD_QUIT: {
        saveColor();
        PostQuitMessage(0);
        break;
	}
    case MOD_SETUP: {
        MessageBox(virtuaWin, configMessage,
				"VirtuaColor " VIRTUACOLOR_VERSION, 0);
        break;
	}
    case MOD_CHANGEDESK:
    case MOD_STEPLEFT:
    case MOD_STEPRIGHT:
    case MOD_STEPUP:
    case MOD_STEPDOWN: {
		DeskID deskid = (DeskID) lParam;
        switchDesk(deskid);
        break;
	}
    case WM_DESTROY: {
        PostQuitMessage(0);
        break;
	}
    default:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    return FALSE;
}

/* ----------------------------------------------------------------------
 * Entry point.
 */
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{
    WNDCLASS wc;
    MSG msg;
    
	/* Initialize application. */
    memset(&wc, 0, sizeof(WNDCLASS));
    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC) MainWndProc;
    wc.hInstance = hInstance;
    wc.lpszClassName = "VirtuaColor.exe";
    if (!RegisterClass(&wc))
		return 0;
    
	/* Create an invisible window to receive messages. */
    if (CreateWindow("VirtuaColor.exe", "VirtuaColor", 
                     WS_POPUP, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0,
                     NULL, NULL, hInstance, NULL) == (HWND)0)
        return 0;
    
	/* Main message loop. */
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return msg.wParam;
}
